/* TAG TIFF high-level glue code.
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1998 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TAG.  TAG is distributed under the terms of the
   TAG License, a copy of which can be found in the TAG distribution; see
   the file LICENSE.

   $Id: tiffglue.c,v 1.2 1998/02/24 22:52:59 tiggr Exp $  */

#include "ttiff-r.h"
#include <tom/util.h>
#include <tiffio.h>

static void
ttiff_error_handler (const char *module, const char *fmt, va_list ap)
{
  char *buf = alloca (8192); /* XXX overrun... */
  sprintf (buf, "%s: ", module);
  vsprintf (buf + strlen (buf), fmt, ap);

  TRT_RAISE (0, c_tom_Conditions_error, byte_string_with_c_string (buf));
}

void
ttiff_initialize (void)
{
  TIFFSetErrorHandler (ttiff_error_handler);
}

tom_object
c_ttiff_ttiff_r_imageRepFromFileNamed_r (tom_object self, selector cmd,
					 tom_object filename)
{
  tom_object r = 0;
  tom_int len;
  TIFF *tif;
  char *s;

  C_STRING_WITH_TOM_STRING (s, len, filename);

  tif = TIFFOpen (s, "r");
  if (!tif)
    return 0;
  {
    char meshed_p;
    tom_int colorspace;
    uint32 im_width, im_height, v;
    uint32 rows_per_strip;
    uint16 bps, spp, sv;
    char **planes;
    char *buf;

    TIFFGetField (tif, TIFFTAG_IMAGEWIDTH, &im_width);
    TIFFGetField (tif, TIFFTAG_IMAGELENGTH, &im_height);
    TIFFGetField (tif, TIFFTAG_BITSPERSAMPLE, &bps);

    TIFFGetField (tif, TIFFTAG_PHOTOMETRIC, &sv);
    if (sv == PHOTOMETRIC_MINISBLACK)
      colorspace = c_tag_Constants_COLOR_SPACE_DEVICE_WHITE;
    else
      TIFFError ("ttiff", "unknown photometric: %d", sv);

    TIFFGetField (tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
    TIFFGetField (tif, TIFFTAG_ROWSPERSTRIP, &rows_per_strip);

    TIFFGetField (tif, TIFFTAG_PLANARCONFIG, &sv);
    meshed_p = sv == PLANARCONFIG_CONTIG;

    buf = _TIFFmalloc (TIFFStripSize (tif));
    if (colorspace == c_tag_Constants_COLOR_SPACE_DEVICE_WHITE)
      {
	int i, in_row, start_row, col, bytes_per_row, spb;
	char *plane, *b, pix, sample_mask;

	/* Grayscale image.  */
	if (spp > 1)
	  TIFFError ("ttiff", "unhandled #samples: %d", spp);

	bytes_per_row =  ((im_width * bps + 7) / 8); 
	plane = xcalloc (im_height, bytes_per_row);
	planes = xmalloc (sizeof *planes);
	planes[0] = plane;
	sample_mask = (1 << bps) - 1;
	spb = 8 / bps;

	/* Meshed or not does not matter for 1 spp.  */
	for (i = start_row = 0; i < TIFFNumberOfStrips (tif);
	     i++, start_row += rows_per_strip)
	  {
	    TIFFReadEncodedStrip (tif, i, buf, -1);

	    for (in_row = 0, b = buf; in_row < rows_per_strip;
		 b += bytes_per_row, in_row++)
	      {
		char *p = &plane[(start_row + in_row) * bytes_per_row];

		if (start_row + in_row == im_height)
		  break;
		for (col = 0; col < im_width; col++)
		  {
		     pix = ((b[col / spb] >> (bps * (spb - 1 - col % spb)))
			    & sample_mask);
		     p[col / spb] |= pix << (bps * (spb - 1 - col % spb));
		  }
	      }
	  }

	r = TRT_SEND (_PI_, CREF (tag_BitmapImageRep), USEL (tom, r_alloc));
	r = TRT_SEND ((void * (*) (void *, void *, void *, tom_float, tom_float, tom_int, tom_byte, tom_byte, tom_int, tom_int)), r,
		      USEL (tag, r_initWithBitmapDataPlanes_p_size__ff__bitsPerSample_i_hasAlpha_o_isPlanar_o_colorSpace_i_bytesPerRow_i),
		      planes, im_width, im_height, bps, 0, !meshed_p,
		      colorspace, bytes_per_row);
      }
    else
      TIFFError ("ttiff", "unhandled colorspace: %d", colorspace);

    _TIFFfree (buf);
  }
  TIFFClose (tif);

  return r;
}
