/* X TAG high-level glue code.
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996-1998 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TAG.  TAG is distributed under the terms of the
   TAG License, a copy of which can be found in the TAG distribution; see
   the file LICENSE.

   $Id: glue.c,v 1.42 1998/04/13 16:44:58 tiggr Exp $  */

#include "x-r.h"
#include <tom/util.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>

/* Iff !0, we've initialized the low-level stuff.  */
static int x_initialized;

static int
x_io_error_handler (Display *d)
{
  TRT_RAISE (0, c_tom_Conditions_error,
	     byte_string_with_c_string ("X IO Error"));
  return 0;
}

static __inline__ void
x_initialize (void)
{
  x_initialized = 1;

  XSetIOErrorHandler (x_io_error_handler);
}

static tom_int
max_mul (tom_int mask, tom_int *pmul)
{
  *pmul = 1;
  while (!(mask & 1))
    mask >>= 1, *pmul <<= 1;

  return mask;
}

static tom_object
color_map_for_screen (tom_object screen, Screen *s)
{
  Display *d = DisplayOfScreen (s);
  Visual *v = DefaultVisualOfScreen (s);
  int i, depth = DefaultDepthOfScreen (s);
  tom_int cmap = DefaultColormapOfScreen (s);
  tom_int rx, bx, gx, rm, gm, bm, bp, cs;
  XStandardColormap cmap_info;
  tom_object o;

  switch (v->class)
    {
    case StaticGray:
      cs = c_tag_Constants_COLOR_SPACE_DEVICE_WHITE;
      rx = bx = gx = rm = gm = bm = bp = 0;
      break;

    case TrueColor:
      if (!v->red_mask || !v->green_mask || !v->blue_mask)
	fatal ("%s: no rgb masks, visual class=%d", __FUNCTION__, v->class);
      cs = c_tag_Constants_COLOR_SPACE_DEVICE_RGB;
      rx = max_mul (v->red_mask, &rm);
      gx = max_mul (v->green_mask, &gm);
      bx = max_mul (v->blue_mask, &bm);
      bp = 0;
      break;

    case PseudoColor:
      i = XGetStandardColormap (d, RootWindowOfScreen (s), &cmap_info,
				XA_RGB_DEFAULT_MAP);
      if (!i)
	fatal ("%s: can't get RGB_DEFAULT_MAP", __FUNCTION__);
#if 0
      if (v->visualid != cmap_info.visualid)
	fatal ("%s: visual id %ld does not match colormap's %ld",
	       __FUNCTION__, (long) v->visualid, (long) cmap_info.visualid);
#endif

      cs = c_tag_Constants_COLOR_SPACE_DEVICE_RGB;
      rx = cmap_info.red_max;
      gx = cmap_info.green_max;
      bx = cmap_info.blue_max;
      rm = cmap_info.red_mult;
      gm = cmap_info.green_mult;
      bm = cmap_info.blue_mult;
      bp = cmap_info.base_pixel;
      cmap = cmap_info.colormap;
      break;

    default:
      fatal ("%s: unhandled visual class %d", __FUNCTION__, v->class);
    }

  o = TRT_SEND (_PI_, CREF (x_Colormap), SEL (r_alloc));
  o = TRT_SEND (MAT (_TP_, _TI_, _TP_, _TI_, _TI_, _TI_,
		     _TI_, _TI_, _TI_, _TI_, _TI_), o,
		SEL (r_init_i_on_r_colorSpace_i_muls__iii__maxes__iii__base_i),
		cmap, screen, cs, rm, gm, bm, rx, gx, bx, bp);
  return o;
}

static __inline__ void
collect_expose_events (Display *d, tom_object wnd, XEvent *e)
{
  typedef struct
  {
    int x, y, w, h;
  } rect;
  Window w = TRT_SEND (, wnd, SEL (i_window));
  int i, c_rects = 128, n_rects = 1;
  XExposeEvent *ee = &e->xexpose;
  rect r[c_rects];

  r[0].x = ee->x;
  r[0].y = ee->y;
  r[0].w = ee->width;
  r[0].h = ee->height;

  while (XCheckWindowEvent (d, w, ExposureMask, e))
    {
      int i;

      for (i = 0; i < n_rects; i++)
	{
	  char xe = r[i].x == ee->x;
	  char ye = r[i].y == ee->y;
	  char we = r[i].w == ee->width;
	  char he = r[i].h == ee->height;

	  if (he && ye)
	    {
	      if (r[i].x + r[i].w == ee->x)
		{
		  /* fprintf (stderr, "side-right\n"); */
		  r[i].w += ee->width;
		  break;
		}
	    }
	  else if (xe && we)
	    {
	      if (r[i].y + r[i].h == ee->y)
		{
		  /* fprintf (stderr, "side-below\n"); */
		  r[i].h += ee->height;
		  break;
		}
	    }
	}
      if (i == n_rects)
	{
	  /* XXX */
	  if (n_rects == c_rects)
	    abort ();

	  n_rects++;
	  r[i].x = ee->x;
	  r[i].y = ee->y;
	  r[i].w = ee->width;
	  r[i].h = ee->height;
	}
    }

  for (i = 0; i < n_rects; i++)
    TRT_SEND ((void (*) (tom_object, selector, tom_float, tom_float,
			 tom_float, tom_float, tom_int)), wnd,
	      SEL (v_expose__ffff__left_i), r[i].x, r[i].y, r[i].w, r[i].h, 0);
}

static __inline__ void
simple_button_event (int what, struct _es_i_tag_Display *tag,
		     tom_object wnd, XEvent *e)
{
  if (!tag->e_mouse)
    tag->e_mouse = (void *) TRT_SEND (_PI_, _mr_c_tag_MouseEvent,
				      SEL (r_alloc));
  if (!wnd)
    fprintf (stderr, "%s: No window %ld for event %d\n",
	     __FUNCTION__, e->xany.window, e->type);
  else
    {
      tom_int mods = 0;

      if (e->xbutton.state & ShiftMask)
	mods |= c_tag_Constants_EVENT_MOD_SHIFT;
      if (e->xbutton.state & ControlMask)
	mods |= c_tag_Constants_EVENT_MOD_CONTROL;
      if (e->xbutton.state & Mod1Mask)
	mods |= c_tag_Constants_EVENT_MOD_COMMAND;
      if (e->xbutton.state & Mod2Mask)
	mods |= c_tag_Constants_EVENT_MOD_ALTERNATE;

      tag->e_mouse = (void *)
	TRT_SEND ((void * (*) (void *, selector, tom_int, void *, tom_int,
			       tom_int, tom_float, tom_float)),
		  tag->e_mouse,
	      USEL (tag, r_init_i_window_r_modifiers_i_buttons_i_location__ff_),
		  what, wnd, mods, 0, e->xbutton.x, e->xbutton.y);
      tag->event = (void *) tag->e_mouse;
      TRT_SEND (, wnd, SEL (v_mouseEvent_r), tag->event);
    }
}

/******************** Display ********************/

void
c_x_Display_v_closeDisplay_p (tom_object self, selector cmd, Display *d)
{
  XCloseDisplay (d);
}

void
i_x_Display_v_internalDispatch (tom_object self, selector cmd)
{
  struct _es_i_x_Display *this = trt_ext_address (self, _ei_i_x_Display);
  struct _es_i_tag_Display *tag = trt_ext_address (self, _ei_i_tag_Display);
  Display *d = this->display;
  int do_unlock = 0, left;
  tom_object wnd;
  XEvent e;

  if (tag->lock)
    {
      do_unlock = 1;
      TRT_SEND (, tag->lock, USEL (tom, v_lock));
    }

  for (left = XEventsQueued (d, QueuedAfterReading);
       left; left = XEventsQueued (d, QueuedAlready))
    {
      XNextEvent (d, &e);

      tag->event = 0;

      wnd = TRT_SEND ((tom_object (*) (void *, selector, tom_int)),
		      this->windows, SEL (r_at_i), e.xany.window);
      if (!wnd)
	fprintf (stderr, __FUNCTION__ ": event %d for unknown window %ld\n",
		 e.type, e.xany.window);
      else switch (e.type)
	{
	case Expose:
	  collect_expose_events (d, wnd, &e);
	  break;

	case GraphicsExpose:
	  TRT_SEND ((void (*) (tom_object, selector, tom_float, tom_float,
			       tom_float, tom_float, tom_int)), wnd,
		    SEL (v_expose__ffff__left_i), e.xgraphicsexpose.x,
		    e.xgraphicsexpose.y, e.xgraphicsexpose.width,
		    e.xgraphicsexpose.height, 0);
	  break;

	case ButtonPress:
	  simple_button_event (c_tag_Constants_EVENT_MOUSE_DOWN, tag, wnd, &e);
	  break;

	case ButtonRelease:
	  simple_button_event (c_tag_Constants_EVENT_MOUSE_UP, tag, wnd, &e);
	  break;

	case MotionNotify:
	  while (XCheckWindowEvent (d, e.xany.window, ButtonMotionMask, &e))
	    ;

	  simple_button_event (c_tag_Constants_EVENT_MOUSE_DRAG, tag, wnd, &e);
	  break;
	  
	case ConfigureNotify:
	  TRT_SEND (, wnd, SEL (v_configureNotify_p), &e);
	  break;

	case UnmapNotify:
	case MapNotify:
	case ReparentNotify:
	case NoExpose:
	  /* Ignore these.  */
	  break;

	default:
	  fprintf (stderr, "%s: %ld: unknown event type %d\n",
		   __FUNCTION__, e.xany.window, e.type);
	}
    }

  if (tag->lock && do_unlock)
    TRT_SEND (, tag->lock, USEL (tom, v_unlock));
}

tom_int
i_x_Display_i_eventsQueued_i (tom_object self, selector cmd, tom_int when)
{
  struct _es_i_x_Display *this = trt_ext_address (self, _ei_i_x_Display);
  Display *d = this->display;

  return XEventsQueued (d, (when == c_x_Constants_QUEUED_AFTER_FLUSH
			    ? QueuedAfterFlush
			    : when == c_x_Constants_QUEUED_AFTER_READING
			    ? QueuedAfterReading
			    : QueuedAlready));
}

tom_object
i_x_Display_r_do_init_r (tom_object self, selector cmd, tom_object display_name)
{
  struct _es_i_x_Display *this = trt_ext_address (self, _ei_i_x_Display);
  struct _es_i_tom_Descriptor *that
    = trt_ext_address (self, _ei_i_tom_Descriptor);
  Display *d;
  tom_int l;
  char *s;

  if (!x_initialized)
    x_initialize ();

  this->name = trt_assign_object_var (self, display_name);

  s = TRT_SEND (_PI_, display_name, USEL (tom, _pi__byteStringContents), &l);
  s = memcpy (alloca (l + 1), s, l);
  s[l] = 0;

  d = XOpenDisplay (s);
  if (!d)
    return 0;

  this->display = d;
  that->descriptor = ConnectionNumber (d);

  return self;
}

tom_object
i_x_Display_r_screens (tom_object self, selector cmd)
{
  struct _es_i_x_Display *this = trt_ext_address (self, _ei_i_x_Display);
  Display *d;
  int i, n;

  if (this->screens)
    return (tom_object) this->screens;

  d = this->display;
  n = ScreenCount (d);
  this->screens = TRT_SEND (_PI_, _mr_c_tom_MutableObjectArray,
			    USEL (tom, r_withCapacity_i), (tom_int) n);

  for (i = 0; i < n; i++)
    {
      Screen *s = ScreenOfDisplay (d, i);
      tom_object os = TRT_SEND (_PI_, _mr_c_x_Screen, SEL (r_alloc));
      os = TRT_SEND (_PI_, os, SEL (r_init_p_on_r), s, self);
      TRT_SEND (, this->screens, SEL (v_add_r), os);
    }

  return (tom_object) this->screens;
}

#define TAG2X_X(DISPLAY, X)  \
  ((X) + (DISPLAY)->bl_x)

#define TAG2X_Y(DISPLAY, Y)  \
  ((DISPLAY)->bl_y - (Y))

void
i_x_Display_v_drawImage_r_at__ff_ (tom_object self, selector cmd, tom_object r,
				   tom_float x1, tom_float y1)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  int pixmap = TRT_SEND (_II_, r, SEL (i_pixmap));
  Screen *s = x->locked_screen;
  Display *d = x->display;

  x1 = TAG2X_X (x, x1);
  y1 = TAG2X_Y (x, y1);

  if (pixmap)
    {
      tom_float w, h;
      XGCValues v;
      int iw, ih;
      GC gc;

      w = TRT_SEND (_FI_, r, SEL (_ff__size), &h);
      iw = w, ih = h;

      v.tile = pixmap;
      v.fill_style = FillTiled;
      v.ts_x_origin = x1;
      v.ts_y_origin = y1 - ih + 1;
      gc = XCreateGC (x->display, x->locked_window,
		      (GCTile | GCFillStyle | GCTileStipXOrigin
		       | GCTileStipYOrigin), &v);
      /* Inherit clipping information.  */
      XCopyGC (x->display, DefaultGCOfScreen (s),
	       GCClipXOrigin | GCClipYOrigin | GCClipMask, gc);
      XFillRectangle (x->display, x->locked_window, gc,
		      x1, y1 - ih + 1, iw, ih);
      XFreeGC (x->display, gc);
    }
  else
    {
      XImage *xim = TRT_SEND (_PI_, r, SEL (p_ximage));

      XPutImage (d, x->locked_window, DefaultGCOfScreen (s), xim,
		 0, 0, x1, y1 - xim->height + 1, xim->width, xim->height);
    }
}

void
i_x_Display_v_drawLine__ffff_ (tom_object self, selector cmd, tom_float x1,
			       tom_float y1, tom_float x2, tom_float y2)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  Screen *s = x->locked_screen;
  Display *d = x->display;

  x1 = TAG2X_X (x, x1);
  x2 = TAG2X_X (x, x2);
  y1 = TAG2X_Y (x, y1);
  y2 = TAG2X_Y (x, y2);

  XDrawLine (d, x->locked_window, DefaultGCOfScreen (s), x1, y1, x2, y2);
}

void
i_x_Display_v_fillRect__ffff_ (tom_object self, selector cmd, tom_float x1,
			       tom_float y1, tom_float w, tom_float h)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  Screen *s = x->locked_screen;
  Display *d = x->display;

  x1 = TAG2X_X (x, x1);
  y1 = TAG2X_Y (x, y1);

  XFillRectangle (d, x->locked_window, DefaultGCOfScreen (s),
		  x1, y1 - h + 1, w, h);
}

void *
i_x_Display_p_loadFont_r (tom_object self, selector cmd, tom_object name)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  Display *d = x->display;
  XFontStruct *xf;
  tom_int len;
  char *s;

  s = TRT_SEND (_PI_, name, USEL (tom, _pi__byteStringContents), &len);
  s = memcpy (alloca (len + 1), s, len);
  s[len] = 0;

  xf = XLoadQueryFont (d, s);
  if (!xf)
    fprintf (stderr, "%s: could not load `%s'\n", __FUNCTION__, s);

  return xf;
}

void
i_x_Display_v_setFont_r (tom_object self, selector cmd, tom_object font)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  XFontStruct *f = TRT_SEND (_PI_, font, SEL (p_font));
  Screen *s = x->locked_screen;
  XGCValues attr;

  attr.font = f->fid;
  XChangeGC (x->display, DefaultGCOfScreen (s), GCFont, &attr);

  x->current_font = trt_assign_object_var (self, font);
}

void
i_x_Display_v_setXColor_r (tom_object self, selector cmd, tom_object xcolor)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  tom_int pixel = TRT_SEND (, xcolor, SEL (i_pixelValue));
  Screen *s = x->locked_screen;
  XGCValues attr;

  attr.foreground = attr.background = pixel;
  XChangeGC (x->display, DefaultGCOfScreen (s),
	     GCForeground | GCBackground, &attr);
}

void
i_x_Display_v_show_r_at__ff_ (tom_object self, selector cmd, tom_object text,
			      tom_float x1, tom_float y1)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  XFontStruct *f = TRT_SEND (_PI_, x->current_font, SEL (p_font));
  Screen *scr = x->locked_screen;
  Display *d = x->display;
  tom_int len;
  char *s;

  x1 = TAG2X_X (x, x1);
  y1 = TAG2X_Y (x, y1);
  y1 -= f->descent;

  /* XXX Uni-flawed.  */
  s = TRT_SEND (_PI_, text, USEL (tom, _pi__byteStringContents), &len);
  XDrawString (d, x->locked_window, DefaultGCOfScreen (scr), x1, y1, s, len);
}

void
i_x_Display_v_clipToRect__ffff_ (tom_object self, selector cmd, tom_float x1,
				 tom_float y1, tom_float w, tom_float h)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  Screen *s = x->locked_screen;
  Display *d = x->display;
  XRectangle rect;
  Region r;

  x1 = TAG2X_X (x, x1);
  y1 = TAG2X_Y (x, y1);

  r = XCreateRegion ();

  rect.x = x1;
  rect.y = y1 - h + 1;
  rect.width = w;
  rect.height = h;

  XUnionRectWithRegion (&rect, r, r);
  XSetRegion (d, DefaultGCOfScreen (s), r);
  XDestroyRegion (r);
  XSetClipOrigin (d, DefaultGCOfScreen (s), 0, 0);
}

void
i_x_Display_v_unclip (tom_object self, selector cmd)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  Screen *s = x->locked_screen;
  Display *d = x->display;

  XSetClipMask (d, DefaultGCOfScreen (s), None);
}

void
i_x_Display_v_copyRect__ffff__to__ff_
  (tom_object self, selector cmd, tom_float x1, tom_float y1,
   tom_float w, tom_float h, tom_float x2, tom_float y2)
{
  struct _es_i_x_Display *x = trt_ext_address (self, _ei_i_x_Display);
  Screen *s = x->locked_screen;
  Display *d = x->display;
  XRectangle rect;
  Region r;

  x1 = TAG2X_X (x, x1);
  y1 = TAG2X_Y (x, y1);
  x2 = TAG2X_X (x, x2);
  y2 = TAG2X_Y (x, y2);

  XCopyArea (d, x->locked_window, x->locked_window, DefaultGCOfScreen (s),
	     x1, y1 - h + 1, w, h, x2, y2 - h + 1);
}

/******************** Font ********************/

tom_float
i_x_Font_f_ascent (tom_object self, selector cmd)
{
  struct _es_i_x_Font *this = trt_ext_address (self, _ei_i_x_Font);
  XFontStruct *font = this->font;
  
  return font->ascent;
}

tom_float
i_x_Font_f_descent (tom_object self, selector cmd)
{
  struct _es_i_x_Font *this = trt_ext_address (self, _ei_i_x_Font);
  XFontStruct *font = this->font;
  
  return font->descent;
}

tom_float
i_x_Font_f_height (tom_object self, selector cmd)
{
  struct _es_i_x_Font *this = trt_ext_address (self, _ei_i_x_Font);
  XFontStruct *font = this->font;
  
  return font->ascent + font->descent;
}

tom_float
i_x_Font_f_widthOf_r (tom_object self, selector cmd, tom_object string)
{
  struct _es_i_x_Font *this = trt_ext_address (self, _ei_i_x_Font);
  XFontStruct *font = this->font;
  XCharStruct cs;
  int a, b, c;
  tom_int len;
  char *s;

  /* XXX Uni-flawed.  */
  s = TRT_SEND (_PI_, string, USEL (tom, _pi__byteStringContents), &len);

  XTextExtents (font, s, len, &a, &b, &c, &cs);

  return cs.width;
}

/******************** CachedImageRep ********************/

/* Get bits from a bitmap where samples retrieved never cross a byte
   boundary.  */
#define CHAR_BITMAP_GET_BITS(BM, BO, NB)  \
  ((((unsigned char *) BM)[(BO) / 8]  \
    >> (8 - (NB) - ((BO) % (8))))  \
   & ((1 << (NB)) - 1))

tom_object
c_x_CachedImageRep_r_withBitmapImageRep_r_onScreen_r
  (tom_object self, selector cmd, tom_object bm, tom_object screen)
{
  char **planes = TRT_SEND (_PI_, bm, USEL (tag, p_planes));
  Screen *s = TRT_SEND (_PI_, screen, SEL (p_screen));
  XPixmapFormatValues *formats;
  tom_object cmap, cim;
  int num_formats;
  XImage *im = 0;
  tom_int csp, ocsp;

  formats = XListPixmapFormats (DisplayOfScreen (s), &num_formats);

  csp = TRT_SEND (, bm, SEL (i_color_space));
  cmap = TRT_SEND (_PI_, screen, SEL (r_colormap));
  ocsp = TRT_SEND (, cmap, SEL (i_color_space));

  if (csp == c_tag_Constants_COLOR_SPACE_DEVICE_WHITE)
    {
      int ibpp = TRT_SEND (, bm, USEL (tag, i_bits_per_pixel));
      int ibpr = TRT_SEND (, bm, USEL (tag, i_bytes_per_row));
      int depth = DefaultDepthOfScreen (s), obpp = depth;
      int bit_padding = BitmapPad (DisplayOfScreen (s));
      int i, w, h, *data, obpr;
      char *plane = planes[0];
      tom_float fw, fh;
      int ix, iy;

      /* Find a fitting OBPP.  */
      for (i = 0; i < num_formats; i++)
	if (formats[i].bits_per_pixel >= obpp)
	  {
	    obpp = formats[i].bits_per_pixel;
	    break;
	  }

      fw = TRT_SEND (_FI_, bm, SEL (_ff__size), &fh);
      w = fw, h = fh;

      obpr = ((w * obpp + bit_padding - 1) / bit_padding) * bit_padding;
      data = xcalloc (1, (obpr / 8) * h);
      im = XCreateImage (DisplayOfScreen (s), DefaultVisualOfScreen (s),
			 depth, ZPixmap, 0, (void *) data, w, h,
			 bit_padding, obpr / 8);
      if (!im)
	fatal ("%s: XCreateImage (ZPixmap) failed", __FUNCTION__);

      if (ibpp == 2)
	{
	  tom_int pixels[4];

	  pixels[0] = TRT_SEND (_II_, cmap, SEL (_iiii__whitePixels),
				pixels + 1, pixels + 2, pixels + 3);

	  for (iy = 0; iy < h; iy++, plane += ibpr)
	    for (ix = 0; ix < w; ix++)
	      XPutPixel (im, ix, iy,
			 pixels[CHAR_BITMAP_GET_BITS (plane, ix * ibpp, ibpp)]);
	}
      else if (ocsp == c_tag_Constants_COLOR_SPACE_DEVICE_RGB)
	{
	  tom_int red_mul, green_mul, blue_mul;
	  tom_int red_max, green_max, blue_max;
	  tom_int base_pixel;

	  base_pixel = TRT_SEND (_II_, cmap, SEL (_iiiiiii__rgbVitals),
				 &red_max, &green_max, &blue_max,
				 &red_mul, &green_mul, &blue_mul);

	  for (iy = 0; iy < h; iy++, plane += ibpr)
	    for (ix = 0; ix < w; ix++)
	      {
		unsigned int pixel
		  = CHAR_BITMAP_GET_BITS (plane, ix * ibpp, ibpp);
		pixel = ((pixel * red_max / ((1 << ibpp) - 1)) * red_mul
			 + (pixel * green_max / ((1 << ibpp) - 1)) * green_mul
			 + (pixel * blue_max / ((1 << ibpp) - 1)) * blue_mul
			 + base_pixel);

		XPutPixel (im, ix, iy, pixel);
	      }
	}
      else
	fatal ("%s: unhandled output color space %d", __FUNCTION__, ocsp);
    }
  else
    fatal ("%s: unhandled bitmap color space %d", __FUNCTION__, csp);

  XFree (formats);

  cim = TRT_SEND (_PI_, self, SEL (r_alloc));
  cim = TRT_SEND (_PI_, cim, SEL (r_init__rpi_),
		  screen, im, (tom_int) 0);
  TRT_SEND (MAT (_TP_, _TF_, _TF_), cim,
	    USEL (tag, v_setSize__ff_), im->width, im->height);

  return cim;
}

tom_object
c_x_CachedImageRep_r_withSize__ff__onScreen_r
  (tom_object self, selector cmd, tom_float w, tom_float h, tom_object screen)
{
  Screen *s = TRT_SEND (_PI_, screen, SEL (p_screen));
  tom_object o;
  Pixmap px;

  px = XCreatePixmap (DisplayOfScreen (s), RootWindowOfScreen (s),
		      w, h, DefaultDepthOfScreen (s));
  if (!px)
    abort ();

  o = TRT_SEND (_PI_, self, SEL (r_alloc));
  o = TRT_SEND ((void * (*) (void *, void *, tom_object, void *, tom_int)), o,
		 SEL (r_init__rpi_), screen, 0, px);
  return o;
}

tom_object
i_x_CachedImageRep_v_dealloc (tom_object self, selector cmd)
{
  struct _es_i_x_CachedImageRep *this
    = trt_ext_address (self, _ei_i_x_CachedImageRep);
  Screen *s = TRT_SEND (_PI_, this->screen, SEL (p_screen));

  if (this->ximage)
    {
      XImage *xim = this->ximage;
      XDestroyImage (xim);
      this->ximage = 0;
    }

  if (this->pixmap)
    {
      XFreePixmap (DisplayOfScreen (s), this->pixmap);
      this->pixmap = 0;
    }
}

tom_int
i_x_CachedImageRep_i_pixmap (tom_object self, selector cmd)
{
  struct _es_i_x_CachedImageRep *this
    = trt_ext_address (self, _ei_i_x_CachedImageRep);

  /* XXX Only create a pixmap if it is interesting to do so?  */
  if (!this->pixmap)
    {
      Screen *s = TRT_SEND (_PI_, this->screen, SEL (p_screen));
      Display *d = DisplayOfScreen (s);
      tom_object td = TRT_SEND (_PI_, this->screen, SEL (r_display));
      int lw = TRT_SEND (_II_, td, SEL (i_locked_window));
      XImage *xim = this->ximage;
      Pixmap px
	= XCreatePixmap (DisplayOfScreen (s), RootWindowOfScreen (s),
			 xim->width, xim->height, DefaultDepthOfScreen (s));
      GC gc = XCreateGC (d, lw, 0 , NULL);
      XPutImage (d, px, gc, xim, 0, 0, 0, 0, xim->width, xim->height);
      XFreeGC (d, gc);
      this->pixmap = px;
    }
  return this->pixmap;
}

/******************** Screen ********************/

tom_object
i_x_Screen_r_init_p_on_r (tom_object self, selector cmd,
			  Screen *s, tom_object display)
{
  struct _es_i_x_Screen *x = trt_ext_address (self, _ei_i_x_Screen);
  struct _es_i_tag_Screen *tag = trt_ext_address (self, _ei_i_tag_Screen);

  tag->display = trt_assign_object_var (self, display);
  x->screen = s;

  x->colormap = trt_assign_object_var (self, color_map_for_screen (self, s));
  x->device = trt_assign_object_var
    (self, TRT_SEND
     ((tom_object (*) (void *, void *, tom_int, tom_float, tom_float,
		       tom_float, tom_float, tom_float, tom_float,
		       tom_int, tom_int)),
      _mr_c_tag_DeviceDescription,
      USEL (tag, r_on_i_resolution__ff__frame__ffff__colorSpace_i_bitsPerSample_i),
      c_tag_Constants_DEVICE_TYPE_SCREEN,
      WidthOfScreen (s) * 25.4 / WidthMMOfScreen (s),
      HeightOfScreen (s) * 25.4 / HeightMMOfScreen (s),
      0.0, 0.0, WidthOfScreen (s), HeightOfScreen (s),
      c_tag_Constants_COLOR_SPACE_DEVICE_RGB, 
      /* XXX 4???  Mon Mar  2 22:07:16 1998, tiggr@gerbil.org  */
      4));

  return self;
}

/******************** Window ********************/

void
i_x_Window_v_configureNotify_p (tom_object self, selector cmd,
				XConfigureEvent *e)
{
  struct _es_i_tag_Window *tag = trt_ext_address (self, _ei_i_tag_Window);
  struct _es_i_x_Window *this = trt_ext_address (self, _ei_i_x_Window);
  int moved, sized;
  tom_float sh;

  TRT_SEND (, this->screen, SEL (_ff__size), &sh);
  e->y = sh - e->y;

  /* XXX Actually, if the size did not change, the move is valid.
     Otherwise, i.e. if we were resized, the position is the position
     relative to our parent (the window manager's window).  */
  sized = tag->f_w != e->width || tag->f_h != e->height;
  moved = tag->f_x != e->x || tag->f_y != e->y;

  TRT_SEND ((void (*) (tom_object, selector, tom_float, tom_float, tom_float,
		       tom_float)), self, USEL (tag, v_noteFrame__ffff_),
	    e->x, e->y - e->height, e->width, e->height);
}

tom_int
c_x_Window_i_createWindow__ffff__with__ii__on_r
  (tom_object self, selector cmd, tom_float x, tom_float y, tom_float w,
   tom_float h, tom_int style, tom_int backing, tom_object screen)
{
  tom_object display = TRT_SEND (_PI_, screen, SEL (r_display));
  Display *d = TRT_SEND (_PI_, display, SEL (p_display));
  Screen *s = TRT_SEND (_PI_, screen, SEL (p_screen));
  XSizeHints hints = {USPosition | USSize, x, y, w, h};
  XSetWindowAttributes attrs;
  unsigned long attr_mask = 0;
  Window wnd;

  /* If enabled, we do not get any ornaments or other garbage.  */
  if (style == c_tag_Constants_WINDOW_STYLE_PLAIN)
    {
      attrs.override_redirect = 1;
      attr_mask |= CWOverrideRedirect;
    }

  wnd = XCreateWindow (d, RootWindowOfScreen (s),
		       x, HeightOfScreen (s) - y - h, w, h, 1,
		       DefaultDepthOfScreen (s), InputOutput,
		       DefaultVisualOfScreen (s), attr_mask, &attrs);

  XSelectInput (d, wnd, (ExposureMask | StructureNotifyMask | ButtonPressMask
			 | ButtonReleaseMask | ButtonMotionMask));
  XSetNormalHints (d, wnd, &hints);
  XMapWindow (d, wnd);

  return wnd;
}

void
c_x_Window_v_setBackground_r_ofWindow_i_onDisplay_p
  (tom_object self, selector cmd,
   tom_object xcolor, tom_int window, void *display)
{
  XSetWindowAttributes attr;

  attr.background_pixel = TRT_SEND (, xcolor, SEL (i_pixelValue));

  XChangeWindowAttributes (display, window, CWBackPixel, &attr);
}

/******************** XColor ********************/

tom_object
i_x_XColor_r_initFor_r_inMap_r (tom_object self, selector cmd,
				tom_object c, tom_object colormap)
{
  struct _es_i_x_XColor *this = trt_ext_address (self, _ei_i_x_XColor);
  int colorspace, i, ir, ig, ib;
  tom_object screen;
  tom_float r, g, b;
  Colormap cmap;
  Display *d;
  Screen *s;
  XColor *xc;

  this->colormap = trt_assign_object_var (self, colormap);

  colorspace = TRT_SEND (, c, USEL (tag, i_colorSpace));
  switch (colorspace)
    {
    case c_tag_Constants_COLOR_SPACE_DEVICE_WHITE:
      r = TRT_SEND (_FI_, c, USEL (tag, f_whiteComponents));
      g = b = r;
      break;

    case c_tag_Constants_COLOR_SPACE_DEVICE_RGB:
      r = TRT_SEND (_FI_, c, USEL (tag, _fff__rgbComponents), &g, &b);
      break;

    default:
      fatal ("%s: unknown colorspace %d", __FUNCTION__, colorspace);
      return 0;
    }

  xc = this->xcolor = xcalloc (1, sizeof (*xc));
  xc->red = ir = 65535 * r;
  xc->green = ig = 65535 * g;
  xc->blue = ib = 65535 * b;

  screen = TRT_SEND (_PI_, colormap, SEL (r_screen));
  s = TRT_SEND (_PI_, screen, SEL (p_screen));
  d = DisplayOfScreen (s);
  cmap = TRT_SEND (, colormap, SEL (i_colormap));
  i = XAllocColor (d, cmap, xc);

  if (ir != xc->red || ig != xc->green || ib != xc->blue)
    fprintf (stderr, "%d: wanted %04x %04x %04x, got %04x %04x %04x (%ld)\n",
	     i, ir, ig, ib, xc->red, xc->green, xc->blue, xc->pixel);

  return self;
}

tom_int
i_x_XColor_i_pixelValue (tom_object self, selector cmd)
{
  struct _es_i_x_XColor *this = trt_ext_address (self, _ei_i_x_XColor);
  XColor *xc = this->xcolor;

  return xc->pixel;
}

void
i_x_Subpath_v_stroke__pi__offsets__ff_ (tom_object self, selector cmd,
					Display *display, tom_int drawable,
					tom_float blx, tom_float bly)
{
  struct _es_i_x_Subpath *this = trt_ext_address (self, _ei_i_x_Subpath);
  float *coords = this->coordinates;
  XPoint points[this->coords_num];
  int i;

  /* Must have been flattened before stroked.  */
  if (this->num_curves)
    abort ();

#undef TAG2X_X
#define TAG2X_X(X, BLX)  ((X) + (BLX))
#undef TAG2X_Y
#define TAG2X_Y(Y, BLY)  ((BLY) - (Y))

  for (i = 0; i < this->coords_num; i++)
    {
      points[i].x = TAG2X_X (coords[2 * i], blx);
      points[i].y = TAG2X_Y (coords[2 * i + 1], bly);
    }

  XDrawLines (display, drawable,
	      DefaultGCOfScreen (DefaultScreenOfDisplay (display)),
	      points, this->coords_num, CoordModeOrigin);

#undef TAG2X_X
#undef TAG2X_Y
}
